package org.owasp.csrf;

import java.io.IOException;
import java.io.OutputStream;
import java.util.Set;
import java.security.NoSuchAlgorithmException;

import javax.servlet.Filter;
import javax.servlet.FilterChain;
import javax.servlet.FilterConfig;
import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.http.Cookie;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;

import org.htmlparser.Node;
import org.htmlparser.Parser;
import org.htmlparser.nodes.TagNode;
import org.htmlparser.tags.FormTag;
import org.htmlparser.tags.InputTag;
import org.htmlparser.tags.LinkTag;
import org.htmlparser.tags.ScriptTag;
import org.htmlparser.util.NodeIterator;
import org.htmlparser.util.NodeList;
import org.htmlparser.util.ParserException;
import org.owasp.csrf.util.CryptoUtil;
import org.owasp.csrf.util.PageUtil;
import org.owasp.csrf.http.MutableHttpResponse;

public class CSRFGuard implements Filter
{
	private final static String CSRFGUARD_CONFIG = "CSRFGuard.xml";
	
	private final static String OWASP_CSRFTOKEN = "__OWASP_CSRFTOKEN";
	
	private ServletContext context = null;
	
	private String errorPage = null;
	
	private Set<String> pages = null;
	
	public void init(FilterConfig config)
	{
		context = config.getServletContext();
		
		errorPage = config.getInitParameter("error-page");
		
		if(errorPage == null) context.log("the CSRFGuard error-page parameter was not properly defined");
		
		String configFile = context.getRealPath("WEB-INF") + "/" + CSRFGUARD_CONFIG;
		
		try
		{
			pages = PageUtil.GetCSRFPages(configFile);
		}
		catch (Exception e)
		{
			context.log(e.getMessage(), e);
		}
	}
	
	public void doFilter(ServletRequest request, ServletResponse response, FilterChain chain)
	{
		if(request instanceof HttpServletRequest && response instanceof HttpServletResponse)
		{
			HttpServletRequest hRequest = (HttpServletRequest)request;
			MutableHttpResponse mResponse = new MutableHttpResponse((HttpServletResponse)response);
			String uri = hRequest.getRequestURI();
			
			try
			{
				if(isValidURI(uri)) { doRequestEnforce(hRequest); }
				
				chain.doFilter(hRequest, mResponse);
				
				if(isValidURI(uri)) { doResponseEnforce(hRequest, mResponse); }
				
				doResponse(hRequest, mResponse);
			}
			catch(CSRFException e)
			{
				doError(hRequest, mResponse);
				
				context.log(e.getMessage(), e);
			}
			catch (Throwable t)
			{
				context.log(t.getMessage(), t);
			}
		}
	}
	
	public void destroy()
	{
		
	}
	
	private boolean isValidURI(String s)
	{
		return pages.contains(s);
	}
	
    private void doError(HttpServletRequest request, MutableHttpResponse response)
    {
    	HttpSession session = request.getSession(false);
    	
    	if(session != null) { session.invalidate(); }
    	
    	try
    	{
    		Cookie c = new Cookie("JSESSIONID", "");
    		c.setMaxAge(-1);
    		
    		response.addCookie(c);
    		response.sendRedirect(errorPage);
    	}
    	catch(Exception e)
    	{
    		context.log(e.getMessage(), e);
    	}
    }
	
    private void doResponse(HttpServletRequest request, MutableHttpResponse response) throws IOException, ServletException
    {
        HttpServletResponse hResponse = response.getResponse();
        OutputStream out = hResponse.getOutputStream();
        byte[] result = response.getContent();
        
        response.setContentLength(result.length);
        out.write(result);
        out.close();
    }
    
    private void doRequestEnforce(HttpServletRequest request) throws CSRFException, NoSuchAlgorithmException
    {
    	HttpSession session = request.getSession(true);
    	String oToken = (String)session.getAttribute(OWASP_CSRFTOKEN);
    	String nToken = (String)request.getParameter(OWASP_CSRFTOKEN);
    	
    	if(oToken == null)
    	{
    		setToken(session);
    	}
    	else if(!oToken.equals(nToken))
    	{
    		throw new CSRFException("caught CSRF attack (URI - " + request.getRequestURI() + " IP - " + request.getRemoteAddr() + " USER - " + request.getRemoteUser() + ")");
    	}
    }
    
	private void setToken(HttpSession session) throws NoSuchAlgorithmException
	{
		String token = CryptoUtil.GenerateRandomToken().replace('+', '_');
		session.setAttribute(OWASP_CSRFTOKEN, token);
	}
	
    private void doResponseEnforce(HttpServletRequest request, MutableHttpResponse response) throws IOException, ServletException
    {
    	HttpSession session = request.getSession(true);
    	String oToken = (String)session.getAttribute(OWASP_CSRFTOKEN);
    	
    	doToken(response, oToken);
    }
    
	private void doToken(MutableHttpResponse response, String oToken)
	{
		try
		{
			Parser p = new Parser();
			
			p.setResource(new String(response.getContent()));
			
			NodeList nl = p.parse(null);
			NodeIterator itr = nl.elements();
			
			while(itr.hasMoreNodes())
			{
				Node n = (Node)itr.nextNode();
				processNodes(n, oToken);
			}
			
			response.setContent(nl.toHtml());
		}
		catch (Exception e)
		{
			context.log(e.getMessage(), e);
		}
	}
	
	private void processNodes(Node node, String oToken) throws ParserException
	{
		if(node instanceof LinkTag)
		{
			LinkTag link = (LinkTag)node;
			
			if(link.isHTTPLink() || link.isHTTPSLink())
			{
				String text = link.getAttribute("href");
				
				if(text.indexOf("?") != -1)
				{
					text = text.replace("?", "?" + OWASP_CSRFTOKEN + "=" + oToken + "&");
				}
				else
				{
					text = text + "?" + OWASP_CSRFTOKEN + "=" + oToken;
				}
				
				link.setLink(text);
			}
		}
		else if(node instanceof FormTag)
		{
			FormTag ft = (FormTag)node;
			InputTag it = new InputTag();
			
			it.setAttribute("type", "hidden");
			it.setAttribute("name", OWASP_CSRFTOKEN);
			it.setAttribute("value", oToken);
			
			NodeList children = ft.getChildren();
			
			children.add(it);
			
			ft.setChildren(children);
		}
		else if(node instanceof ScriptTag)
		{
			/** FIXME: we need to modify JavaScript that do document.url or document.open type calls **/
		}
		else if(node instanceof TagNode)
		{
	         TagNode tag = (TagNode)node;
	         NodeList nl = tag.getChildren();
	         
	         if(nl != null)
	         {
	        	 for(NodeIterator itr = nl.elements(); itr.hasMoreNodes(); processNodes(itr.nextNode(), oToken));
	         }
		}
	}
}
